@@ -1,4 +1,4 @@
-# Copyright (c) 1997-2000 Graham Barr <gbarr@pobox.com>. All rights reserved.
+# Copyright (c) 1997-2004 Graham Barr <gbarr@pobox.com>. All rights reserved.
 # This program is free software; you can redistribute it and/or
 # modify it under the same terms as Perl itself.
 
@@ -7,6 +7,7 @@ package Net::LDAP;
 use strict;
 use IO::Socket;
 use IO::Select;
+use Tie::Hash;
 use vars qw($VERSION $LDAP_VERSION @ISA);
 use Convert::ASN1 qw(asn_read);
 use Net::LDAP::Message;
@@ -20,11 +21,15 @@ use Net::LDAP::Constant qw(LDAP_SUCCESS
 			   LDAP_LOCAL_ERROR
 			   LDAP_PARAM_ERROR
 			   LDAP_INAPPROPRIATE_AUTH
+			   LDAP_SERVER_DOWN
+			   LDAP_USER_CANCELED
+			   LDAP_EXTENSION_START_TLS
+			   LDAP_UNAVAILABLE
 			);
 
-$VERSION 	= 0.25;
-@ISA     	= qw(Net::LDAP::Extra);
-$LDAP_VERSION 	= 2;      # default LDAP protocol version
+$VERSION 	= "0.31";
+@ISA     	= qw(Tie::StdHash Net::LDAP::Extra);
+$LDAP_VERSION 	= 3;      # default LDAP protocol version
 
 # Net::LDAP::Extra will only exist is someone use's the module. But we need
 # to ensure the package stash exists or perl will complain that we inherit
@@ -96,9 +101,20 @@ sub new {
   my $arg  = &_options;
   my $obj  = bless {}, $type;
 
-  $obj->_connect($host, $arg) or return;
+  foreach my $uri (ref($host) ? @$host : ($host)) {
+    my $scheme = $arg->{scheme} || 'ldap';
+    (my $h = $uri) =~ s,^(\w+)://,, and $scheme = $1;
+    my $meth = $obj->can("connect_$scheme") or next;
+    $h =~ s,/.*,,; # remove path part
+    $h =~ s/%([A-Fa-f0-9]{2})/chr(hex($1))/eg; # unescape
+    if (&$meth($obj, $h, $arg)) {
+      $obj->{net_ldap_uri} = $uri;
+      last;
+    }
+  }
+
+  return undef unless $obj->{net_ldap_socket};
 
-  $obj->{net_ldap_host}    = $host;
   $obj->{net_ldap_resp}    = {};
   $obj->{net_ldap_version} = $arg->{version} || $LDAP_VERSION;
   $obj->{net_ldap_async}   = $arg->{async} ? 1 : 0;
@@ -110,20 +126,100 @@ sub new {
 
   $obj->debug($arg->{debug} || 0 );
 
-  $obj;
+  $obj->outer;
 }
 
-sub _connect {
+sub connect_ldap {
   my ($ldap, $host, $arg) = @_;
 
   $ldap->{net_ldap_socket} = IO::Socket::INET->new(
-    PeerAddr => $host,
-    PeerPort => $arg->{port} || '389',
-    Proto    => 'tcp',
+    PeerAddr   => $host,
+    PeerPort   => $arg->{port} || '389',
+    Proto      => 'tcp',
+    MultiHomed => $arg->{multihomed},
+    Timeout    => defined $arg->{timeout}
+		 ? $arg->{timeout}
+		 : 120
+  ) or return undef;
+  
+  $ldap->{net_ldap_host} = $host;
+}
+
+
+# Different OpenSSL verify modes.
+my %ssl_verify = qw(none 0 optional 1 require 3);
+
+sub connect_ldaps {
+  my ($ldap, $host, $arg) = @_;
+  require IO::Socket::SSL;
+
+  $ldap->{'net_ldap_socket'} = IO::Socket::SSL->new(
+    PeerAddr 	    => $host,
+    PeerPort 	    => $arg->{'port'} || '636',
+    Proto    	    => 'tcp',
+    Timeout  	    => defined $arg->{'timeout'} ? $arg->{'timeout'} : 120,
+    _SSL_context_init_args($arg)
+  ) or return undef;
+
+  $ldap->{net_ldap_host} = $host;
+}
+
+sub _SSL_context_init_args {
+  my $arg = shift;
+
+  my $verify = 0;
+  my ($clientcert,$clientkey,$passwdcb);
+
+  if (exists $arg->{'verify'}) {
+      my $v = lc $arg->{'verify'};
+      $verify = 0 + (exists $ssl_verify{$v} ? $ssl_verify{$v} : $verify);
+  }
+
+  if (exists $arg->{'clientcert'}) {
+      $clientcert = $arg->{'clientcert'};
+      if (exists $arg->{'clientkey'}) {
+	  $clientkey = $arg->{'clientkey'};
+      } else {
+	  require Carp;
+	  Carp::croak("Setting client public key but not client private key");
+      }
+  }
+
+  if (exists $arg->{'keydecrypt'}) {
+      $passwdcb = $arg->{'keydecrypt'};
+  }
+
+  (
+    SSL_cipher_list => defined $arg->{'ciphers'} ? $arg->{'ciphers'} : 'ALL',
+    SSL_ca_file     => exists  $arg->{'cafile'}  ? $arg->{'cafile'}  : '',
+    SSL_ca_path     => exists  $arg->{'capath'}  ? $arg->{'capath'}  : '',
+    SSL_key_file    => $clientcert ? $clientkey : undef,
+    SSL_passwd_cb   => $passwdcb,
+    SSL_use_cert    => $clientcert ? 1 : 0,
+    SSL_cert_file   => $clientcert,
+    SSL_verify_mode => $verify,
+    SSL_version     => defined $arg->{'sslversion'} ? $arg->{'sslversion'} :
+                       'sslv2/3',
+  );
+}
+
+sub connect_ldapi {
+  my ($ldap, $peer, $arg) = @_;
+
+  $peer = $ENV{LDAPI_SOCK} || "/var/lib/ldapi"
+    unless length $peer;
+
+  require IO::Socket::UNIX;
+
+  $ldap->{net_ldap_socket} = IO::Socket::UNIX->new(
+    Peer => $peer,
     Timeout  => defined $arg->{timeout}
 		 ? $arg->{timeout}
 		 : 120
-  );
+  ) or return undef;
+
+  $ldap->{net_ldap_host} = 'localhost';
+  $ldap->{net_ldap_peer} = $peer;
 }
 
 sub message {
@@ -233,15 +329,29 @@ sub bind {
 
     my $sasl = $passwd;
     # Tell the SASL object our user identifier
-    $sasl->user("dn: $dn") unless $sasl->user;
+    $sasl->callback( user => "dn: $stash{name}")
+      unless $sasl->callback('user');
+
+    my $sasl_conn = $sasl->client_new("ldap",$ldap->{net_ldap_host});
+
+    # Tell SASL the local and server IP addresses
+    $sasl_conn->property(
+      sockname => $ldap->{net_ldap_socket}->sockname,
+      peername => $ldap->{net_ldap_socket}->peername,
+    );
+
+    my $initial = $sasl_conn->client_start;
+
+    return _error($ldap, $mesg, LDAP_LOCAL_ERROR, "$@")
+      unless defined($initial);
 
     $passwd = {
-      mechanism   => $sasl->name,
-      credentials => $sasl->initial
+      mechanism   => $sasl_conn->mechanism,
+      credentials => (length($initial) ? $initial : undef)
     };
 
     # Save data, we will need it later
-    $mesg->_sasl_info($stash{name},$control,$sasl);
+    $mesg->_sasl_info($stash{name},$control,$sasl_conn);
   }
 
   $stash{authentication} = { $auth_type => $passwd };
@@ -410,7 +520,7 @@ sub modify {
 	    operation => $opcode,
 	    modification => {
 	      type => $attr,
-	      vals => $val
+	      vals => ref($val) ? $val : [$val]
 	    }
 	  };
 	}
@@ -566,7 +676,7 @@ sub extension {
     if $ldap->{net_ldap_version} < 3;
 
   $mesg->encode(
-    extendedRequest => {
+    extendedReq => {
       requestName  => $arg->{name},
       requestValue => $arg->{value}
     },
@@ -590,6 +700,11 @@ sub sync {
   $err;
 }
 
+sub disconnect {
+  my $self = shift;
+  _drop_conn($self, LDAP_USER_CANCELED, "Explicit disconnect");
+}
+
 sub _sendmesg {
   my $ldap = shift;
   my $mesg = shift;
@@ -606,7 +721,10 @@ sub _sendmesg {
       if $debug & 4;
   }
 
-  syswrite($ldap->socket, $mesg->pdu, length($mesg->pdu))
+  my $socket = $ldap->socket
+      or return _error($ldap, $mesg, LDAP_SERVER_DOWN, "$!");
+
+  syswrite($socket, $mesg->pdu, length($mesg->pdu))
     or return _error($ldap, $mesg, LDAP_LOCAL_ERROR,"$!");
 
   # for CLDAP, here we need to recode when we were sent
@@ -633,14 +751,14 @@ sub _sendmesg {
 sub _recvresp {
   my $ldap = shift;
   my $what = shift;
-  my $sock = $ldap->socket;
+  my $sock = $ldap->socket or return LDAP_SERVER_DOWN;
   my $sel = IO::Select->new($sock);
   my $ready;
 
   for( $ready = 1 ; $ready ; $ready = $sel->can_read(0)) {
     my $pdu;
     asn_read($sock, $pdu)
-      or return LDAP_OPERATIONS_ERROR;
+      or return _drop_conn($ldap, LDAP_OPERATIONS_ERROR, "Communications Error");
 
     my $debug;
     if ($debug = $ldap->debug) {
@@ -657,13 +775,20 @@ sub _recvresp {
     my $result = $LDAPResponse->decode($pdu)
       or return LDAP_DECODING_ERROR;
 
-    my $mid = $result->{messageID};
+    my $mid  = $result->{messageID};
+    my $mesg = $ldap->{net_ldap_mesg}->{$mid};
 
-    my $mesg = $ldap->{net_ldap_mesg}->{$mid} or
-      do {
-	print STDERR "Unexpected PDU, ignored\n" if $debug & 10;
-	next;
-      };
+    unless ($mesg) {
+      if (my $ext = $result->{protocolOp}{extendedResp}) {
+	if (($ext->{responseName} || '') eq '1.3.6.1.4.1.1466.20036') {
+	  # notice of disconnection
+	  return _drop_conn($ldap, LDAP_SERVER_DOWN, "Notice of Disconnection");
+	}
+      }
+
+      print STDERR "Unexpected PDU, ignored\n" if $debug & 10;
+      next;
+    }
 
     $mesg->decode($result) or
       return $mesg->code;
@@ -677,6 +802,22 @@ sub _recvresp {
   return LDAP_SUCCESS;
 }
 
+sub _drop_conn {
+  my ($self, $err, $etxt) = @_;
+
+  my $sock = delete $self->{net_ldap_socket};
+  close($sock) if $sock;
+
+  if (my $msgs = delete $self->{net_ldap_mesg}) {
+    foreach my $mesg (values %$msgs) {
+      $mesg->set_error($err, $etxt);
+    }
+  }
+
+  $err;
+}
+
+
 sub _forgetmesg {
   my $ldap = shift;
   my $mesg = shift;
@@ -738,6 +879,7 @@ sub schema {
     : Net::LDAP::Schema->new($mesg->entry);
 }
 
+
 sub root_dse {
   my $ldap = shift;
   my %arg  = @_;
@@ -750,6 +892,9 @@ sub root_dse {
 		  supportedSASLMechanisms
 		  supportedLDAPVersion
 		)];
+  my $root = $arg{attrs} && $ldap->{net_ldap_root_dse};
+
+  return $root if $root;
 
   my $mesg = $ldap->search(
     base   => '',
@@ -758,7 +903,13 @@ sub root_dse {
     attrs  => $attrs,
   );
 
-  $mesg->entry;
+  require Net::LDAP::RootDSE;
+  $root = $mesg->entry;
+  bless $root, 'Net::LDAP::RootDSE' if $root; # Naughty, but there you go :-)
+
+  $ldap->{net_ldap_root_dse} = $root unless $arg{attrs};
+
+  return $root;
 }
 
 sub start_tls {
@@ -766,6 +917,7 @@ sub start_tls {
   my $arg  = &_options;
   my $sock = $ldap->socket;
 
+  require IO::Socket::SSL;
   require Net::LDAP::Extension;
   my $mesg = $ldap->message('Net::LDAP::Extension' => $arg);
 
@@ -777,7 +929,7 @@ sub start_tls {
 
   $mesg->encode(
     extendedReq => {
-      requestName => "1.3.6.1.4.1.1466.20037",
+      requestName => LDAP_EXTENSION_START_TLS,
     }
   );
 
@@ -787,12 +939,23 @@ sub start_tls {
   return $mesg
     if $mesg->code;
 
-  require Net::LDAPS;
+  delete $ldap->{net_ldap_root_dse};
+
   $arg->{sslversion} = 'tlsv1' unless defined $arg->{sslversion};
-  IO::Socket::SSL::context_init( { Net::LDAPS::SSL_context_init_args($arg) } );
-  (IO::Socket::SSL::socketToSSL($sock) and tie *{$sock}, 'IO::Socket::SSL', $sock)
-    ? $mesg
-    : _error($ldap, $mesg, LDAP_OPERATIONS_ERROR, $@);
+  IO::Socket::SSL::context_init( { _SSL_context_init_args($arg) } );
+  my $sock_class = ref($sock);
+
+  return $mesg
+    if IO::Socket::SSL::socketToSSL($sock, {_SSL_context_init_args($arg)});
+
+  my $err = $@;
+
+  if ($sock_class ne ref($sock)) {
+    $err = $sock->errstr;
+    bless $sock, $sock_class;
+  }
+
+  _error($ldap, $mesg, LDAP_OPERATIONS_ERROR, $err);
 }
 
 sub cipher {
@@ -818,5 +981,29 @@ sub version {
     : $ldap->{net_ldap_version};
 }
 
+sub outer {
+  my $self = shift;
+  return $self if tied(%$self);
+  my %outer;
+  tie %outer, ref($self), $self;
+  ++$self->{net_ldap_refcnt};
+  bless \%outer, ref($self);
+}
+
+sub inner {
+  tied(%{$_[0]}) || $_[0];
+}
+
+sub TIEHASH {
+  $_[1];
+}
+
+sub DESTROY {
+  my $ldap = shift;
+  my $inner = tied(%$ldap) or return;
+  _drop_conn($inner, LDAP_UNAVAILABLE, "Implicit disconnect")
+    unless --$inner->{net_ldap_refcnt};
+}
+
 1;
 
